@@ -19,20 +19,16 @@ TODO:
   * integrate screen dpi w/ ppi and text
 
 """
-from __future__ import (absolute_import, division, print_function,
-                        unicode_literals)
-
-import six
-
-import threading
+try:
+    import threading
+except ImportError:
+    import dummy_threading as threading
 import numpy as np
 from collections import OrderedDict
 from math import radians, cos, sin
 from matplotlib import cbook, rcParams, __version__
 from matplotlib.backend_bases import (
-    _Backend, FigureCanvasBase, FigureManagerBase, RendererBase, cursors)
-from matplotlib.cbook import maxdict
-from matplotlib.figure import Figure
+    _Backend, FigureCanvasBase, FigureManagerBase, RendererBase)
 from matplotlib.font_manager import findfont, get_font
 from matplotlib.ft2font import (LOAD_FORCE_AUTOHINT, LOAD_NO_HINTING,
                                 LOAD_DEFAULT, LOAD_NO_AUTOHINT)
@@ -42,13 +38,11 @@ from matplotlib.transforms import Bbox, BboxBase
 from matplotlib import colors as mcolors
 
 from matplotlib.backends._backend_agg import RendererAgg as _RendererAgg
-from matplotlib import _png
 
-try:
+from matplotlib.backend_bases import _has_pil
+
+if _has_pil:
     from PIL import Image
-    _has_pil = True
-except ImportError:
-    _has_pil = False
 
 backend_version = 'v2.2'
 
@@ -70,11 +64,6 @@ class RendererAgg(RendererBase):
     context instance that controls the colors/styles
     """
 
-    @property
-    @cbook.deprecated("2.2")
-    def debug(self):
-        return 1
-
     # we want to cache the fonts at the class level so that when
     # multiple figures are created we can reuse them.  This helps with
     # a bug on windows where the creation of too many figures leads to
@@ -110,25 +99,13 @@ class RendererAgg(RendererBase):
     def __setstate__(self, state):
         self.__init__(state['width'], state['height'], state['dpi'])
 
-    def _get_hinting_flag(self):
-        if rcParams['text.hinting']:
-            return LOAD_FORCE_AUTOHINT
-        else:
-            return LOAD_NO_HINTING
-
-    # for filtering to work with rasterization, methods needs to be wrapped.
-    # maybe there is better way to do it.
-    def draw_markers(self, *kl, **kw):
-        return self._renderer.draw_markers(*kl, **kw)
-
-    def draw_path_collection(self, *kl, **kw):
-        return self._renderer.draw_path_collection(*kl, **kw)
-
     def _update_methods(self):
-        self.draw_quad_mesh = self._renderer.draw_quad_mesh
         self.draw_gouraud_triangle = self._renderer.draw_gouraud_triangle
         self.draw_gouraud_triangles = self._renderer.draw_gouraud_triangles
         self.draw_image = self._renderer.draw_image
+        self.draw_markers = self._renderer.draw_markers
+        self.draw_path_collection = self._renderer.draw_path_collection
+        self.draw_quad_mesh = self._renderer.draw_quad_mesh
         self.copy_from_bbox = self._renderer.copy_from_bbox
         self.get_content_extents = self._renderer.get_content_extents
 
@@ -140,10 +117,8 @@ class RendererAgg(RendererBase):
         return np.array(region), extents
 
     def draw_path(self, gc, path, transform, rgbFace=None):
-        """
-        Draw the path
-        """
-        nmax = rcParams['agg.path.chunksize'] # here at least for testing
+        # docstring inherited
+        nmax = rcParams['agg.path.chunksize']  # here at least for testing
         npts = path.vertices.shape[0]
 
         if (nmax > 100 and npts > nmax and path.should_simplify and
@@ -173,7 +148,6 @@ class RendererAgg(RendererBase):
                 raise OverflowError("Exceeded cell block limit (set "
                                     "'agg.path.chunksize' rcparam)")
 
-
     def draw_mathtext(self, gc, x, y, s, prop, angle):
         """
         Draw the math text using matplotlib.mathtext
@@ -188,9 +162,8 @@ class RendererAgg(RendererBase):
         self._renderer.draw_text_image(font_image, x, y + 1, angle, gc)
 
     def draw_text(self, gc, x, y, s, prop, angle, ismath=False, mtext=None):
-        """
-        Render the text
-        """
+        # docstring inherited
+
         if ismath:
             return self.draw_mathtext(gc, x, y, s, prop, angle)
 
@@ -218,14 +191,10 @@ class RendererAgg(RendererBase):
             font, np.round(x - xd + xo), np.round(y + yd + yo) + 1, angle, gc)
 
     def get_text_width_height_descent(self, s, prop, ismath):
-        """
-        Get the width, height, and descent (offset from the bottom
-        to the baseline), in display coords, of the string *s* with
-        :class:`~matplotlib.font_manager.FontProperties` *prop*
-        """
+        # docstring inherited
+
         if ismath in ["TeX", "TeX!"]:
             # todo: handle props
-            size = prop.get_size_in_points()
             texmanager = self.get_texmanager()
             fontsize = prop.get_size_in_points()
             w, h, d = texmanager.get_text_width_height_descent(
@@ -248,6 +217,7 @@ class RendererAgg(RendererBase):
         return w, h, d
 
     def draw_tex(self, gc, x, y, s, prop, angle, ismath='TeX!', mtext=None):
+        # docstring inherited
         # todo, handle props, angle, origins
         size = prop.get_size_in_points()
 
@@ -265,12 +235,12 @@ class RendererAgg(RendererBase):
         self._renderer.draw_text_image(Z, x, y, angle, gc)
 
     def get_canvas_width_height(self):
-        'return the canvas width and height in display coords'
+        # docstring inherited
         return self.width, self.height
 
     def _get_agg_font(self, prop):
         """
-        Get the font for text instance t, cacheing for efficiency
+        Get the font for text instance t, caching for efficiency
         """
         fname = findfont(prop)
         font = get_font(fname)
@@ -282,41 +252,38 @@ class RendererAgg(RendererBase):
         return font
 
     def points_to_pixels(self, points):
-        """
-        convert point measures to pixes using dpi and the pixels per
-        inch of the display
-        """
-        return points*self.dpi/72.0
-
-    def tostring_rgb(self):
-        return self._renderer.tostring_rgb()
-
-    def tostring_argb(self):
-        return self._renderer.tostring_argb()
+        # docstring inherited
+        return points * self.dpi / 72
 
     def buffer_rgba(self):
-        return self._renderer.buffer_rgba()
+        return memoryview(self._renderer)
+
+    def tostring_argb(self):
+        return np.asarray(self._renderer).take([3, 0, 1, 2], axis=2).tobytes()
+
+    def tostring_rgb(self):
+        return np.asarray(self._renderer).take([0, 1, 2], axis=2).tobytes()
 
     def clear(self):
         self._renderer.clear()
 
     def option_image_nocomposite(self):
+        # docstring inherited
+
         # It is generally faster to composite each image directly to
         # the Figure, and there's no file size benefit to compositing
         # with the Agg backend
         return True
 
     def option_scale_image(self):
-        """
-        agg backend doesn't support arbitrary scaling of image.
-        """
+        # docstring inherited
         return False
 
     def restore_region(self, region, bbox=None, xy=None):
         """
         Restore the saved region. If bbox (instance of BboxBase, or
         its extents) is given, only the region specified by the bbox
-        will be restored. *xy* (a tuple of two floasts) optionally
+        will be restored. *xy* (a pair of floats) optionally
         specifies the new position (the LLC of the original region,
         not the LLC of the bbox) where the region will be restored.
 
@@ -374,28 +341,22 @@ class RendererAgg(RendererBase):
         post_processing is plotted (using draw_image) on it.
         """
 
-        # WARNING:  For agg_filter to work, the renderer's method need to
-        # overridden in the class. See draw_markers and draw_path_collections.
-
         width, height = int(self.width), int(self.height)
 
-        buffer, bounds = self.tostring_rgba_minimized()
-
-        l, b, w, h = bounds
+        buffer, (l, b, w, h) = self.tostring_rgba_minimized()
 
         self._renderer = self._filter_renderers.pop()
         self._update_methods()
 
         if w > 0 and h > 0:
-            img = np.fromstring(buffer, np.uint8)
+            img = np.frombuffer(buffer, np.uint8)
             img, ox, oy = post_processing(img.reshape((h, w, 4)) / 255.,
                                           self.dpi)
             gc = self.new_gc()
             if img.dtype.kind == 'f':
                 img = np.asarray(img * 255., np.uint8)
             img = img[::-1]
-            self._renderer.draw_image(
-                gc, l + ox, height - b - h + oy, img)
+            self._renderer.draw_image(gc, l + ox, height - b - h + oy, img)
 
 
 class FigureCanvasAgg(FigureCanvasBase):
@@ -420,33 +381,21 @@ class FigureCanvasAgg(FigureCanvasBase):
 
     def draw(self):
         """
-        Draw the figure using the renderer
+        Draw the figure using the renderer.
         """
         self.renderer = self.get_renderer(cleared=True)
-        # acquire a lock on the shared font cache
-        RendererAgg.lock.acquire()
-
-        toolbar = self.toolbar
-        try:
-            # if toolbar:
-            #     toolbar.set_cursor(cursors.WAIT)
+        with RendererAgg.lock:
             self.figure.draw(self.renderer)
             # A GUI class may be need to update a window using this draw, so
             # don't forget to call the superclass.
-            super(FigureCanvasAgg, self).draw()
-        finally:
-            # if toolbar:
-            #     toolbar.set_cursor(toolbar._lastCursor)
-            RendererAgg.lock.release()
+            super().draw()
 
     def get_renderer(self, cleared=False):
         l, b, w, h = self.figure.bbox.bounds
         key = w, h, self.figure.dpi
-        try: self._lastKey, self.renderer
-        except AttributeError: need_new_renderer = True
-        else:  need_new_renderer = (self._lastKey != key)
-
-        if need_new_renderer:
+        reuse_renderer = (hasattr(self, "renderer")
+                          and getattr(self, "_lastKey", None) == key)
+        if not reuse_renderer:
             self.renderer = RendererAgg(w, h, self.figure.dpi)
             self._lastKey = key
         elif cleared:
@@ -454,7 +403,7 @@ class FigureCanvasAgg(FigureCanvasBase):
         return self.renderer
 
     def tostring_rgb(self):
-        '''Get the image as an RGB byte string
+        """Get the image as an RGB byte string.
 
         `draw` must be called at least once before this function will work and
         to update the renderer for any subsequent changes to the Figure.
@@ -462,11 +411,11 @@ class FigureCanvasAgg(FigureCanvasBase):
         Returns
         -------
         bytes
-        '''
+        """
         return self.renderer.tostring_rgb()
 
     def tostring_argb(self):
-        '''Get the image as an ARGB byte string
+        """Get the image as an ARGB byte string.
 
         `draw` must be called at least once before this function will work and
         to update the renderer for any subsequent changes to the Figure.
@@ -474,85 +423,144 @@ class FigureCanvasAgg(FigureCanvasBase):
         Returns
         -------
         bytes
-
-        '''
+        """
         return self.renderer.tostring_argb()
 
     def buffer_rgba(self):
-        '''Get the image as an RGBA byte string
+        """Get the image as a memoryview to the renderer's buffer.
 
         `draw` must be called at least once before this function will work and
         to update the renderer for any subsequent changes to the Figure.
 
         Returns
         -------
-        bytes
-        '''
+        memoryview
+        """
         return self.renderer.buffer_rgba()
 
     def print_raw(self, filename_or_obj, *args, **kwargs):
         FigureCanvasAgg.draw(self)
         renderer = self.get_renderer()
-        original_dpi = renderer.dpi
-        renderer.dpi = self.figure.dpi
-        if isinstance(filename_or_obj, six.string_types):
-            fileobj = open(filename_or_obj, 'wb')
-            close = True
-        else:
-            fileobj = filename_or_obj
-            close = False
-        try:
-            fileobj.write(renderer._renderer.buffer_rgba())
-        finally:
-            if close:
-                fileobj.close()
-            renderer.dpi = original_dpi
+        with cbook._setattr_cm(renderer, dpi=self.figure.dpi), \
+                cbook.open_file_cm(filename_or_obj, "wb") as fh:
+            fh.write(renderer._renderer.buffer_rgba())
+
     print_rgba = print_raw
 
-    def print_png(self, filename_or_obj, *args, **kwargs):
-        FigureCanvasAgg.draw(self)
-        renderer = self.get_renderer()
-        original_dpi = renderer.dpi
-        renderer.dpi = self.figure.dpi
+    def print_png(self, filename_or_obj, *args,
+                  metadata=None, pil_kwargs=None,
+                  **kwargs):
+        """
+        Write the figure to a PNG file.
 
-        version_str = 'matplotlib version ' + __version__ + \
-            ', http://matplotlib.org/'
-        metadata = OrderedDict({'Software': version_str})
-        user_metadata = kwargs.pop("metadata", None)
-        if user_metadata is not None:
-            metadata.update(user_metadata)
+        Parameters
+        ----------
+        filename_or_obj : str or PathLike or file-like object
+            The file to write to.
 
-        try:
-            with cbook.open_file_cm(filename_or_obj, "wb") as fh:
+        metadata : dict, optional
+            Metadata in the PNG file as key-value pairs of bytes or latin-1
+            encodable strings.
+            According to the PNG specification, keys must be shorter than 79
+            chars.
+
+            The `PNG specification`_ defines some common keywords that may be
+            used as appropriate:
+
+            - Title: Short (one line) title or caption for image.
+            - Author: Name of image's creator.
+            - Description: Description of image (possibly long).
+            - Copyright: Copyright notice.
+            - Creation Time: Time of original image creation
+              (usually RFC 1123 format).
+            - Software: Software used to create the image.
+            - Disclaimer: Legal disclaimer.
+            - Warning: Warning of nature of content.
+            - Source: Device used to create the image.
+            - Comment: Miscellaneous comment;
+              conversion from other image format.
+
+            Other keywords may be invented for other purposes.
+
+            If 'Software' is not given, an autogenerated value for matplotlib
+            will be used.
+
+            For more details see the `PNG specification`_.
+
+            .. _PNG specification: \
+                https://www.w3.org/TR/2003/REC-PNG-20031110/#11keywords
+
+        pil_kwargs : dict, optional
+            If set to a non-None value, use Pillow to save the figure instead
+            of Matplotlib's builtin PNG support, and pass these keyword
+            arguments to `PIL.Image.save`.
+
+            If the 'pnginfo' key is present, it completely overrides
+            *metadata*, including the default 'Software' key.
+        """
+        from matplotlib import _png
+
+        if metadata is None:
+            metadata = {}
+        metadata = {
+            "Software":
+                f"matplotlib version{__version__}, http://matplotlib.org/",
+            **metadata,
+        }
+
+        if pil_kwargs is not None:
+            from PIL import Image
+            from PIL.PngImagePlugin import PngInfo
+            buf, size = self.print_to_buffer()
+            # Only use the metadata kwarg if pnginfo is not set, because the
+            # semantics of duplicate keys in pnginfo is unclear.
+            if "pnginfo" not in pil_kwargs:
+                pnginfo = PngInfo()
+                for k, v in metadata.items():
+                    pnginfo.add_text(k, v)
+                pil_kwargs["pnginfo"] = pnginfo
+            pil_kwargs.setdefault("dpi", (self.figure.dpi, self.figure.dpi))
+            (Image.frombuffer("RGBA", size, buf, "raw", "RGBA", 0, 1)
+             .save(filename_or_obj, format="png", **pil_kwargs))
+
+        else:
+            FigureCanvasAgg.draw(self)
+            renderer = self.get_renderer()
+            with cbook._setattr_cm(renderer, dpi=self.figure.dpi), \
+                    cbook.open_file_cm(filename_or_obj, "wb") as fh:
                 _png.write_png(renderer._renderer, fh,
                                self.figure.dpi, metadata=metadata)
-        finally:
-            renderer.dpi = original_dpi
 
     def print_to_buffer(self):
         FigureCanvasAgg.draw(self)
         renderer = self.get_renderer()
-        original_dpi = renderer.dpi
-        renderer.dpi = self.figure.dpi
-        try:
-            result = (renderer._renderer.buffer_rgba(),
-                      (int(renderer.width), int(renderer.height)))
-        finally:
-            renderer.dpi = original_dpi
-        return result
+        with cbook._setattr_cm(renderer, dpi=self.figure.dpi):
+            return (renderer._renderer.buffer_rgba(),
+                    (int(renderer.width), int(renderer.height)))
 
     if _has_pil:
-        # add JPEG support
-        def print_jpg(self, filename_or_obj, *args, **kwargs):
+
+        # Note that these methods should typically be called via savefig() and
+        # print_figure(), and the latter ensures that `self.figure.dpi` already
+        # matches the dpi kwarg (if any).
+
+        def print_jpg(self, filename_or_obj, *args, dryrun=False,
+                      pil_kwargs=None, **kwargs):
             """
+            Write the figure to a JPEG file.
+
+            Parameters
+            ----------
+            filename_or_obj : str or PathLike or file-like object
+                The file to write to.
+
             Other Parameters
             ----------------
             quality : int
-                The image quality, on a scale from 1 (worst) to
-                95 (best). The default is 95, if not given in the
-                matplotlibrc file in the savefig.jpeg_quality parameter.
-                Values above 95 should be avoided; 100 completely
-                disables the JPEG quantization stage.
+                The image quality, on a scale from 1 (worst) to 100 (best).
+                The default is :rc:`savefig.jpeg_quality`.  Values above
+                95 should be avoided; 100 completely disables the JPEG
+                quantization stage.
 
             optimize : bool
                 If present, indicates that the encoder should
@@ -562,37 +570,45 @@ class FigureCanvasAgg(FigureCanvasBase):
             progressive : bool
                 If present, indicates that this image
                 should be stored as a progressive JPEG file.
+
+            pil_kwargs : dict, optional
+                Additional keyword arguments that are passed to
+                `PIL.Image.save` when saving the figure.  These take precedence
+                over *quality*, *optimize* and *progressive*.
             """
             buf, size = self.print_to_buffer()
-            if kwargs.pop("dryrun", False):
+            if dryrun:
                 return
             # The image is "pasted" onto a white background image to safely
             # handle any transparency
             image = Image.frombuffer('RGBA', size, buf, 'raw', 'RGBA', 0, 1)
             rgba = mcolors.to_rgba(rcParams['savefig.facecolor'])
-            color = tuple([int(x * 255.0) for x in rgba[:3]])
+            color = tuple([int(x * 255) for x in rgba[:3]])
             background = Image.new('RGB', size, color)
             background.paste(image, image)
-            options = {k: kwargs[k]
-                       for k in ['quality', 'optimize', 'progressive', 'dpi']
-                       if k in kwargs}
-            options.setdefault('quality', rcParams['savefig.jpeg_quality'])
-            if 'dpi' in options:
-                # Set the same dpi in both x and y directions
-                options['dpi'] = (options['dpi'], options['dpi'])
+            if pil_kwargs is None:
+                pil_kwargs = {}
+            for k in ["quality", "optimize", "progressive"]:
+                if k in kwargs:
+                    pil_kwargs.setdefault(k, kwargs[k])
+            pil_kwargs.setdefault("quality", rcParams["savefig.jpeg_quality"])
+            pil_kwargs.setdefault("dpi", (self.figure.dpi, self.figure.dpi))
+            return background.save(
+                filename_or_obj, format='jpeg', **pil_kwargs)
 
-            return background.save(filename_or_obj, format='jpeg', **options)
         print_jpeg = print_jpg
 
-        # add TIFF support
-        def print_tif(self, filename_or_obj, *args, **kwargs):
+        def print_tif(self, filename_or_obj, *args, dryrun=False,
+                      pil_kwargs=None, **kwargs):
             buf, size = self.print_to_buffer()
-            if kwargs.pop("dryrun", False):
+            if dryrun:
                 return
             image = Image.frombuffer('RGBA', size, buf, 'raw', 'RGBA', 0, 1)
-            dpi = (self.figure.dpi, self.figure.dpi)
-            return image.save(filename_or_obj, format='tiff',
-                              dpi=dpi)
+            if pil_kwargs is None:
+                pil_kwargs = {}
+            pil_kwargs.setdefault("dpi", (self.figure.dpi, self.figure.dpi))
+            return image.save(filename_or_obj, format='tiff', **pil_kwargs)
+
         print_tiff = print_tif
 
 
